/*
 * ====================================================================
 * Copyright (c) 1995-1997 Lyonel VINCENT.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mod_put: implements the PUT and DELETE methods
 *
 * Lyonel VINCENT <vincent@hpwww.ec-lyon.fr>
 * modified by Krischan Jodies <krischan@roko.goe.net> to make it compile for
 * Apache 1.3
 * modified by Carolyn Weiss <cweiss@us.ibm.com> for EBCDIC support
 * modified by Guenter Knauf <info@gknw.de> to make it compile for
 * Apache 2.0
 * modified by Chuck Remes <cremes@opendarwin.org> to support PUT resume
 *
 * usage:
 *	PUT_EnablePut		On|Off
 *		default value:	Off
 *		context:	Directory
 *		effect:		enables (or disables) the PUT method for a
 *				whole directory.
 *
 *	PUT_EnableDelete	On|Off
 *		default value:	Off
 *		context:	Directory
 *		effect:		enables (or disables) the DELETE method for a
 *				whole directory.
 *	PUT_BlockSize		<int value>
 *		default value:	2048
 *		context:	Directory
 *		effect:		sets the apr_size_t for the transfer block size.
 *	PUT_DirMask		<octal value>
 *		default value:	007
 *		context:	Directory
 *		effect:		sets the apr_fileperms_t for the directory created.
 *	PUT_FileMask		<octal value>
 *		default value:	007
 *		context:	Directory
 *		effect:		sets the apr_fileperms_t for the file created.
 *
 * For now only with Apache 2.1.0 and up:
 *
 *	PUT_EnableAppend	On|Off
 *		default value:	Off
 *		context:	Directory
 *		effect:		enables (or disables) PUT resume. Only
 *				in scope when PUT_EnablePut is On.
 *
 */

#define MOD_PUT_ADVERTISE

#define MODULE_RELEASE "2.0.8"

#include "ap_compat.h"
#include "ap_mmn.h"
#include "apr.h"
#include "apr_strings.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#if (MODULE_MAGIC_NUMBER_MAJOR < 20020628)
#error  Apache 2.0.40 required
#endif

#define DEFAULT_BLOCK_SIZE 2048
//#define DEFAULT_FILE_PERMS APR_OS_DEFAULT
#define DEFAULT_FILE_PERMS APR_UREAD|APR_UWRITE
#define DEFAULT_DIR_PERMS APR_UREAD|APR_UWRITE|APR_UEXECUTE

typedef struct
{
  int put_enabled;
  int append_enabled;
  int delete_enabled;
  apr_size_t blocksize;
  apr_fileperms_t dirmask;
  apr_fileperms_t filemask;
} put_config_rec;

module AP_MODULE_DECLARE_DATA put_module;

static void *create_put_dir_config (apr_pool_t *p, char *d)
{
  put_config_rec * sec = (put_config_rec *)apr_pcalloc (p, sizeof(put_config_rec));

  if (!sec) return NULL; /* no memory... */

  sec -> put_enabled = 0;
  sec -> append_enabled = 0;
  sec -> delete_enabled = 0;
  sec -> blocksize = DEFAULT_BLOCK_SIZE;
  sec -> dirmask = DEFAULT_DIR_PERMS;
  sec -> filemask = DEFAULT_FILE_PERMS;
  return sec;
}

static const command_rec put_cmds[] = {
  AP_INIT_FLAG("PUT_EnablePut", ap_set_flag_slot,
              (void *)APR_OFFSETOF(put_config_rec, put_enabled),
              OR_AUTHCFG, "Limited to 'on' or 'off'"),
#if (MODULE_MAGIC_NUMBER_MAJOR >= 20040425)
  AP_INIT_FLAG("PUT_EnableAppend", ap_set_flag_slot,
              (void *) APR_OFFSETOF(put_config_rec, append_enabled),
              OR_AUTHCFG, "Limited to 'on' or 'off'"),
#endif
  AP_INIT_FLAG("PUT_EnableDelete", ap_set_flag_slot,
              (void *)APR_OFFSETOF(put_config_rec, delete_enabled),
              OR_AUTHCFG, "Limited to 'on' or 'off'"),
  AP_INIT_TAKE1("PUT_BlockSize", ap_set_int_slot,
              (void *)APR_OFFSETOF(put_config_rec, blocksize),
              OR_AUTHCFG, "numeric block size"),
  AP_INIT_TAKE1("PUT_DirMask", ap_set_int_slot,
              (void *)APR_OFFSETOF(put_config_rec, dirmask),
              OR_AUTHCFG, "numeric directory mask"),
  AP_INIT_TAKE1("PUT_FileMask", ap_set_int_slot,
              (void *)APR_OFFSETOF(put_config_rec, filemask),
              OR_AUTHCFG, "numeric file mask"),
  { NULL }
};

static void make_dirs(apr_pool_t * p, const char * filename, apr_fileperms_t dmask)
{
  char * sto = apr_pstrdup(p, filename), * slash=sto;

  slash = strrchr(slash, '/');
  if (slash) {
      *slash = '\0';
      apr_dir_make_recursive(sto, DEFAULT_DIR_PERMS & ~dmask, p);
  }
}

#if (MODULE_MAGIC_NUMBER_MAJOR >= 20040425)
/* Parses the HTTP headers to find the Content-Range component. If present and
 * valid, the sender is transmitting a range of bytes. In the context of a
 * M_PUT operation this means a transmission is resuming. For our purposes we
 * don't care what the values are; we only care that the range is valid.
 *
 * One other minor detail. On most platforms the apr_off_t will refer to a
 * 64-bit entity. To properly print this (in error logs or for debugging)
 * remember to use the %lld form to get all 8 bytes to print.
 */
/* Shamelessly stolen from mod_dav.c */
static int parse_range(request_rec *r,
                       apr_off_t *range_start, apr_off_t *range_end)
{
  const char *range_c;
  char *range;
  char *dash;
  char *slash;
  char *errp;

  range_c = apr_table_get(r->headers_in, "content-range");
  if (range_c == NULL)
    return 0;

  range = apr_pstrdup(r->pool, range_c);
  if (strncasecmp(range, "bytes ", 6) != 0
        || (dash = ap_strchr(range, '-')) == NULL
        || (slash = ap_strchr(range, '/')) == NULL) {
   /* malformed header. ignore it (per S14.16 of RFC2616) */
     return 0;
  }

  *dash++ = *slash++ = '\0';

  /* ignore invalid ranges. (per S14.16 of RFC2616) */
  if (apr_strtoff(range_start, range + 6, &errp, 10)
        || *errp || *range_start < 0) {
      return 0;
  }

  if (apr_strtoff(range_end, dash, &errp, 10)
        || *errp || *range_end < 0 || *range_end < *range_start) {
      return 0;
  }

  if (*slash != '*') {
    apr_off_t dummy;

    if (apr_strtoff(&dummy, slash, &errp, 10)
          || *errp || dummy <= *range_end) {
          return 0;
    }
  }

  /* we now have a valid range */
  return 1;
}
#endif

static int do_put(request_rec *r)
{
  put_config_rec *sec = (put_config_rec *)ap_get_module_config (r->per_dir_config, &put_module);
  apr_file_t * f;
  int result=OK;
  char * buffer;
  apr_size_t len=0;
//  apr_fileperms_t perms=sec->filemask;
//  apr_fileperms_t old_umask;
  apr_off_t range_start, range_end; // values are ignored; only need for local storage on the stack
  int has_range = 0;

  if(!sec->put_enabled) return HTTP_FORBIDDEN;

  //old_umask = umask(sec->filemask);
  if(r->path_info)        /* a directory did not exist */
  {
    r->filename = apr_pstrcat(r->pool, r->filename, r->path_info, NULL);
    make_dirs(r->pool, r->filename, sec->dirmask);
  }

#if (MODULE_MAGIC_NUMBER_MAJOR >= 20040425)
  if (sec->append_enabled)
  {
    /* The logic is simple. If this module has "append" enabled we need to determine if the current request
     * is for an append operation or if we are creating something new. If the M_PUT arrives without a
     * Content-Range set in the header then this is a "create" PUT. If it arrives with a valid Content-Range
     * range then this is an "append" PUT. Open the file using the appropriate File Flags depending on
     * the value of has_range.
     *
     * Anyone looking to append will likely do a header-only GET to "stat" this file and receive its current SIZE.
     * Using that value they can being a new PUT operation seeking to a file marker SIZE+1 bytes into the file
     * and set that as the beginning of the content range. This is exactly how FTP RESTART functions.
     */
    has_range = parse_range(r, &range_start, &range_end);

    /* If the range is valid and the open succeeds, continue on.
     * If the range is valid and the open fails, set has_range to false and let the next apr_file_open truncate the file.
     * If the range is invalid, this short circuits the apr_file_open operation and falls through to the open & truncate op.
     */
    if (has_range && apr_file_open(&f, r->filename, APR_WRITE|APR_CREATE|APR_BINARY|APR_APPEND, sec->filemask, r->pool))
    {
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, "can't open the file in append mode; append skipped: %s", r->filename);
      has_range = 0;
    }
  } /* end if sec->append_enabled */
#endif

  /* Open and truncate the file if has_range is false (not an append).
   * If has_range == true, file has been successfully opened already; skip this.
   */
  if (!has_range && apr_file_open(&f, r->filename, APR_WRITE|APR_CREATE|APR_BINARY|APR_TRUNCATE, sec->filemask, r->pool))
  {
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, "file permissions deny server write access: %s", r->filename);
//    umask(old_umask);
//  apr_file_perms_set(r->filename, sec->filemask);
    return HTTP_FORBIDDEN;
  }

  if((result=ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK)) == OK)
  {
    if(ap_should_client_block(r))
    {
      buffer = apr_palloc(r->pool, sec->blocksize);
      while((len=ap_get_client_block(r, buffer, sec->blocksize)) > 0)
        apr_file_write(f, buffer, &len);
    }
    /* does this really still work with 2.0 ?? */
    ap_send_http_header(r);
    ap_rprintf(r, "<HTML>File %s created.</HTML>\n", ap_escape_html(r->pool, r->uri));
  }

  apr_file_close(f);
//  umask(old_umask);
//  apr_file_perms_set(r->filename, sec->filemask);
  return result;
}

static int do_delete(request_rec *r)
{
  put_config_rec *sec = (put_config_rec *)ap_get_module_config (r->per_dir_config, &put_module);

  if(!sec->delete_enabled) return HTTP_FORBIDDEN;
  if(r->finfo.filetype != APR_REG) return HTTP_NOT_FOUND;
  if(apr_stat(&r->finfo, r->filename, APR_FINFO_NORM, r->pool) != APR_SUCCESS) return HTTP_NOT_FOUND;

  if(apr_file_remove(r->filename, r->pool))
  {
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, "file permissions deny file deletion: %s", r->filename);
    return HTTP_FORBIDDEN;
  }

  /* does this really still work with 2.0 ?? */
  ap_send_http_header(r);
  ap_rprintf(r, "<HTML>File %s deleted.</HTML>\n", ap_escape_html(r->pool, r->uri));

  return OK;
}

static int put_handler(request_rec *r)
{
  if(r->proxyreq) return DECLINED;
  if(r->method_number == M_PUT)
  {
#ifdef CHARSET_EBCDIC
    ap_checkconv(r);
#endif
    return do_put(r);
  }
  else
  if(r->method_number == M_DELETE) return do_delete(r);
  return DECLINED;
}

static int put_init_handler(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
#ifdef MOD_PUT_ADVERTISE
  ap_add_version_component(p, "mod_put/" MODULE_RELEASE);
#endif
  return OK;
}

static void put_register_hook(apr_pool_t *p)
{
  ap_hook_post_config(put_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler(put_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA put_module = {
  STANDARD20_MODULE_STUFF,
  create_put_dir_config,	/* create per-directory config structure */
  NULL,				/* merge per-directory config structures */
  NULL,				/* create per-server config structure    */
  NULL,				/* merge per-server config structures    */
  put_cmds,			/* command apr_table_t */
  put_register_hook		/* register hooks */
};


